/*************************************************************************************
GENETSKI ALGORITAM ZA RESAVANJE PROBLEMA STRONG MINIMUM ENERGY TOPOLOGY (SMETP)
ZA SLUCAJ SA SIMETRICNIM PODACIMA

Implementiran je GA za CPU, kao i za GPU (CUDA) arhitekturu. 
Autor: Aleksandar Kartelj, Matematicki Fakultet, Beograd
**************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cutil_inline.h>
#include "ga.cuh"
#include "util.h"
#include "cuPrintf1.cuh"

//informacije o problemu (koristimo konstantnu memoriju, jer je brza od globalne. Puni se na startu
__constant__ int probNn;
__constant__ int probNl;
//Broj linkova u bajtovima, uracunat je simbol za term. nulu ukupno n*/
__constant__ int probNlBytes;
__constant__ link probLink[MAXEDGES];
//lista indeksa suseda: na i-tom mestu pocinju susedi i-tog cvora (neighbourIdx)*/
__constant__  int nIdx[MAXNODES];
__device__ unsigned long int next[POPSIZE];
/*tezina funkcije cilja, napravljene su vestacke funkcije
tako da budu slozenosti objvalhardness*O(n^3), specijalno
ako je vrednost 0, radi se originalna funkcija cilja*/
__constant__ int objvalhardness;
/*gustina grafa, koristi se pri inicijalizaciji
vrednosti. Sto je manje, graf ce biti redji*/
__constant__ float graphdensity;


//glavna funkcija
__host__ void ga(char *path, int randSeed, int hIterations,int objhard, float grphdens)                                    
{
	int i,j;
	//datum i vreme
	time_t t;
	//pocetak izvrs, momenat pronalazenja najbolje, kraj
	clock_t start, found=0,end=0;
	//pracenje najbolje jedinke
	unsigned short int lastBest;
	unsigned short int *cBest;
	unsigned short int hBest;
	//menjamo stepen mutacije, ako se nista ne menja
	int bestRepeated;
	//izmenjeni stepen mutacije, na CPU i na GPU
	float hModifiedMutationProb;
	float *cModifiedMutationProb;
	//putanja izlaznog fajla
	char outPath[255]="./out.dat";
	//privremenom drzimo populaciju ovde dok je ne prepisemo na GPU
	chrom popcurrentTemp[POPSIZE];
	//niz sa seed-ovima slucajnih promenljivih, za svaku nit moramo da
	//imamo poseban seed
	//ovaj je na CPU
	unsigned long int seedsTemp[POPSIZE];
	//ovaj je na GPU
	unsigned long int *seeds;
	//broj iteracija spoljne petlje je ukupan broj iteracija/broj iteracija u samom kernelu
	hIterations/=KERNELLOOP;
	//pokazivac na izlazni fajl
	FILE *out;
	//broj cvorova i grana, lokalne promenljive, za potrebe GPU ove dve info ce stojati u 
	//konstantnoj memoriji
	int nn, nl;
	//populacija na GPU
	chrom *popcurrent;
	/*podaci o vremenu, po koracima*/
	clock_t tL0, tL1;
	//vreme prolaska glavne petlje, cisto da bi se videlo
	//koliki je vremenski znacaj inicijalizacije
	int loopTime=0;
	//kolicina utrosene konstantne memorije
	int constMemSize=0;
	//uzimamo datum i vreme pocetka izvrsavanja
	time(&t);
	//merenje procesorskog vremena pocinje
	start=clock();
	//ucitavanje instance problema iz fajla
	loadInstance(path, &constMemSize);
	//sprintf(outPath,"./tmp_%d_%d.dat",randSeed,num);
	printf("\nSTRONG MINIMUM ENERGY TOPOLOGY PROBLEM (SMETP)\n");
	printf("\nBroj iteracija:  %d\n",hIterations);
	//postavljamo prosledjeni random seed za ceo program (prolsedjeno kroz arg. kom. linije)
	srand(randSeed);
	//pravimo niz seed-ova, za svaki thread posebno
	for(i=0; i<POPSIZE; i++){
		seedsTemp[i]=rand();
	}
	//parametri vestacka tezina funkcije cilja i gustina polaznog grafa
	HANDLE_ERROR(cudaMemcpyToSymbol("objvalhardness", &objhard, sizeof(int)));
	HANDLE_ERROR(cudaMemcpyToSymbol("graphdensity", &grphdens, sizeof(float)));
	//prebacujemo seed-ove na GPU
	HANDLE_ERROR(cudaMalloc((void**)&seeds, sizeof(unsigned long int)*POPSIZE));
	HANDLE_ERROR(cudaMemcpy(seeds, seedsTemp, sizeof(unsigned long int)*POPSIZE, cudaMemcpyHostToDevice));
	//postavljamo r. seed-ove u device memoriji, ovo je dodatni kernel za inicijalizaciju
	k_srand<<<1,1>>>(seeds);
	//nicijalizacija pocetne populacije
	init(popcurrentTemp);
	//kada smo u hostu, cuda constantnu memoriju moramo da ciljamo ovako
	HANDLE_ERROR(cudaMemcpyFromSymbol(&nn, "probNn",sizeof(nn)));
	HANDLE_ERROR(cudaMemcpyFromSymbol(&nl, "probNl",sizeof(nl)));
	//kopiramo sve u GPU (device) memoriju
	HANDLE_ERROR(cudaMalloc((void**)&popcurrent, sizeof(chrom)*POPSIZE));
	HANDLE_ERROR(cudaMemcpy(popcurrent, popcurrentTemp, sizeof(chrom)*POPSIZE, cudaMemcpyHostToDevice));
	//alociramo podatak o najboljoj jedinki koji ce se vracati kroz cuda kernel
	HANDLE_ERROR(cudaMalloc(&cBest, sizeof(unsigned short int)));
	//generisemo najpre random populaciju
	generateRandomPopulation<<<1,POPSIZE>>>(popcurrent);
	evpopInitial<<<1,POPSIZE>>>(popcurrent);
	cudaPrintfDisplay(stdout, false);
	HANDLE_ERROR(cudaMemcpy(&hBest, cBest, sizeof(unsigned short int), cudaMemcpyDeviceToHost));
	//printf("hBest: %d\n",hBest);
	//cudaPrintfDisplay(stdout, false);
	lastBest=INF;
	hBest=INF;
	//alociramo podatak za prenos verovatnoce mutacije na device
	HANDLE_ERROR(cudaMalloc(&cModifiedMutationProb, sizeof(float)));
	HANDLE_ERROR(cudaMemcpy(cModifiedMutationProb, &hModifiedMutationProb, sizeof(float), cudaMemcpyHostToDevice));
	//dodatno merimo i trajanje same petlje, bez prvog dela programa
	tL0=clock();
	for(i=1;i<=hIterations;i++)                            
	{
		//azuiraramo stanje najbolje jedinke i stepen mutacije
		if((MINIMIZATION && hBest<lastBest) || (!MINIMIZATION && hBest>lastBest))
		{
			lastBest=hBest;
			found=clock();
			bestRepeated=1;
			hModifiedMutationProb=MUTATIONPROB;
			HANDLE_ERROR(cudaMemcpy(cModifiedMutationProb, &hModifiedMutationProb, sizeof(float), cudaMemcpyHostToDevice));
		}else{
			bestRepeated++;
			if(bestRepeated%1000==0){
				hModifiedMutationProb+=0.005;
				HANDLE_ERROR(cudaMemcpy(cModifiedMutationProb, &hModifiedMutationProb, sizeof(float), cudaMemcpyHostToDevice));
			}
		}
		printf("Najbolja:\t%d\t= %d\n",lastBest,i*KERNELLOOP); 
		cudaPrintfInit();
	
		//kernel koji u sebi sadrzi sve operatore
		GAallInOneOperator<<<1,POPSIZE>>>(popcurrent,cModifiedMutationProb,cBest);
		//ispis iz CUDA koda
		cudaPrintfDisplay(stdout, false);
		cudaPrintfEnd();
		//prepisujemo najbolju jedinku na CPU
		HANDLE_ERROR(cudaMemcpy(&hBest, cBest, sizeof(unsigned short int), cudaMemcpyDeviceToHost));
		HANDLE_ERROR(cudaMemcpy(&hModifiedMutationProb, cModifiedMutationProb, sizeof(float), cudaMemcpyDeviceToHost));
	}
	//gotovo je izvrsavanje petlje
	tL1=clock();
	loopTime+=(tL1-tL0);
	//kao i celog programa
	end=clock();
	/*zavrsilo se izvrsavanje, ispis...*/
	out=fopen(outPath,"a");
	if(out==NULL){
		fprintf(stderr, "Greska pri otvaranju fajla %s\n", outPath);
		exit(0);
	}
	fprintf(out,"-----------------------------------------------------------\
				\ntime:\t\t %s\
				\npath:\t\t%s\
				\nseed:\t\t%d\
				\nmaxGen:\t%d\
				\npopSize:\t%d\
				\nobjHard:\t%d\
				\nnelite:\t%d\
				\nstartMut:\t%.4f\
				\nendMut:\t%.4f\
				\ncrossProb:\t%.4f\
				\ntournSize:\t%d\
				\ngraphDens:\t%.4f\
				\nbest:\t\t%d\
				\nsec:\t\t%.4f\
				\ntotSec:\t%.4f\
				\nmemory:\
				\n-----------\
				\nconst:\t%d\n",ctime(&t),strrstr(path,"/"),randSeed,hIterations*KERNELLOOP, POPSIZE,objhard,NELITE,MUTATIONPROB,
				hModifiedMutationProb,CROSSOVERPROB,NCOMPET,grphdens,lastBest,
				((float) (found - start)) / CLOCKS_PER_SEC,((float) (end - start)) / CLOCKS_PER_SEC,constMemSize);
	fclose(out);
	//TODO: nisu oslobodjeni popcurrent[tid].lbit nizovi, izbacuje gresku kada se pokusa cudaFree
	cudaFree(popcurrent);
	cudaFree(cModifiedMutationProb);
	cudaFree(cBest);
	cudaFree(seeds);
	cudaFree(nIdx);
	cudaFree(probLink);
}                                                      


__host__ void loadInstance(char *path, int *constMemSize){
	FILE *fp;
	int j,a,b,w;
	int neighbNum=0;
	int probNnTemp;
	int probNlTemp;
	int probNlBytesTemp;
	link *probL;
	int *nIdxTemp;
	size_t nIdxSize;
	size_t probLSize;

	fp=fopen(path,"r");
	if(fp == NULL){
		printf("Ne moze da se otvori fajl %s \n",path);
		exit(0);
	}

	fscanf(fp, "%d %d",&probNnTemp, &probNlTemp);
	/*grane su u fajlu zapisane orijentisano, nas interesuje samo
	simetricni slucaj, tako da ih je duplo manje*/
	probNlTemp=probNlTemp/2;
	probNlBytesTemp=(probNlTemp/EDGEMEMSIZE)+3;
	/*pakujemo indekse susedstva*/
	nIdxSize=sizeof(int)*(probNnTemp-1);
	nIdxTemp=(int*)malloc(nIdxSize);
	if(nIdx==NULL || nIdxTemp==NULL){
		fprintf(stderr, "Greska pri alokaciji nIdx i nIdxTemp");
	}
	neighbNum=probNnTemp-1;
	nIdxTemp[0]=0;
	for(j=1; j<probNnTemp-1; j++){
		nIdxTemp[j]=(nIdxTemp[j-1]+neighbNum);
		neighbNum--;
	}
	//for(j=0; j<probNnTemp-1; j++)
	//	printf("%d=%d\n",j,nIdxTemp[j]);

	probLSize=probNlTemp * sizeof(link);
	probL = (link*)malloc(probLSize);
	if(probL == NULL){ 
		fprintf(stderr, "Greska pri alokaciji linkova");
		exit (0); 
	}
	j=0;
	while(j<probNlTemp){
		fscanf(fp,"%d %d %d",&a,&b,&w); 
		if(a<b){
			probL[j].from=a-1;
			probL[j].to=b-1;
			probL[j].weight=w;
			printf("(%d,%d)=%d\n",probL[j].from, probL[j].to, probL[j].weight);
			j++;
		}
	}
	(*constMemSize)+=sizeof(probNn)+sizeof(probNl)+probLSize+nIdxSize+sizeof(probNlBytes);
	/*kopiramo sada probTemp i nIdxTemp u konstantnu memoriju, mnogo brzu nego sto je globalna*/
	HANDLE_ERROR(cudaMemcpyToSymbol("probNn", &probNnTemp,sizeof(probNn)));
	HANDLE_ERROR(cudaMemcpyToSymbol("probNl", &probNlTemp,sizeof(probNl)));
	HANDLE_ERROR(cudaMemcpyToSymbol("probNlBytes", &probNlBytesTemp,sizeof(probNlBytes)));
	HANDLE_ERROR(cudaMemcpyToSymbol("probLink", probL,probLSize));
	HANDLE_ERROR(cudaMemcpyToSymbol("nIdx", nIdxTemp,nIdxSize));
	/*provera, da'l upisuje u simbol*/
	/*HANDLE_ERROR(cudaMemcpyFromSymbol(&temp, "probNn",sizeof(probNn)));*/
	free(probL);
	free(nIdxTemp);
	fclose(fp);
}

__host__ void init(chrom popcurrentTemp[POPSIZE]){
	int j;
	int nn, nl, nlBytes;
	HANDLE_ERROR(cudaMemcpyFromSymbol(&nn, "probNn",sizeof(nn)));
	HANDLE_ERROR(cudaMemcpyFromSymbol(&nl, "probNl",sizeof(nl)));
	HANDLE_ERROR(cudaMemcpyFromSymbol(&nlBytes, "probNlBytes",sizeof(nl)));
	int memL=sizeof(unsigned char)*nlBytes;
	/*alociramo hromozome*/
	for(j=0;j<POPSIZE;j++)                  
	{
		HANDLE_ERROR(cudaMalloc((void**)&popcurrentTemp[j].lbit, memL));
	}	
	printf("Alocirano %d karaktera za svaki niz grana\n",memL);
}

__device__ void generateRandom(unsigned char *lbit, int tid){
	float random;
	int i;
	/*inicijalizujemo linkove (slucajno)*/
	for(i=0;i<probNl;i++)         
	{
		/*inicijalizujemo slucajno 0 ili 1 vrednosti*/
		random=k_rand(tid)*1.0/RAND_MAX; 
		if(random<graphdensity){
			setBit(i,lbit,1);
		}
		else{
			setBit(i,lbit,0);
		}
	}
}

__global__ void generateRandomPopulation(chrom *popcurrent){
	int tid=threadIdx.x + blockIdx.x * blockDim.x;
	if(tid<POPSIZE)
		generateRandom(popcurrent[tid].lbit,tid);

}

/*evaluacija populacije, poziva se samo na pocetku*/
__global__ void evpopInitial(chrom *popcurrent) 
{
	int tid=threadIdx.x;
	__shared__ unsigned  short int sfit[POPSIZE];
	__shared__ unsigned char slbit[POPSIZE][EDGEMEMNUM];	

	sfit[tid]=popcurrent[tid].fit;
	for(int i=0; i<POPSIZE; i++){
		slbit[tid][i]=popcurrent[tid].lbit[i];
	}
	__syncthreads();
	sfit[tid]=fitness(slbit[tid],tid);
	__syncthreads();
	bitonicSort(tid, sfit, slbit);
	__syncthreads();
	replaceEqual(tid,sfit,slbit);
	__syncthreads();
	bitonicSort(tid, sfit, slbit);
	__syncthreads();
	//cuPrintf("%d => %d\n",tid,sfit[tid]);
}                               


__device__ unsigned short int fitness(unsigned char *lbit, int tid)   
{
	unsigned short int sum=0;
	short int z[MAXNODES];
	int i;
	if(Pvalid(lbit,objvalhardness, tid)){
		for(i=0; i<probNn; i++){
			z[i]=0;
		}
		for(i=0; i<probNl; i++){
			/*ako je grana pripadajuca, azuriramo energiju senzora
			na toj poziciji (u slucaju da je veca*/
			if(getBit(i,lbit)==1){
				/*mora da se azurira sa obe strane grane*/
				if(probLink[i].weight>z[probLink[i].from])
					z[probLink[i].from]=probLink[i].weight;
				if(probLink[i].weight>z[probLink[i].to])
					z[probLink[i].to]=probLink[i].weight;
			}
		}
		//if(tid==0){
		for(i=0; i<probNn; i++){
	//		cuPrintf("z[%d]=%d; ",i,z[i]);
			sum+=z[i];
		}
	//		cuPrintf("Sum=%d\n",sum);
		//}
		return sum;
	}else{
		return INF;
	}
}                             

__device__ int edgeIndex(int i,int j){
	int tidx;
	if(i<j)
		tidx=nIdx[i]+(j-i)-1;
	else
		tidx=nIdx[j]+(i-j)-1;
	if(tidx<probNl)
		return tidx;
	else
		return -1;
}

__device__ int bfs(unsigned char *lbit, int tid){
	int visited[MAXNODES];
	int stack[MAXNODES];
	int sp=0;
	int i;
	int eidx;
	int hasNext=1;
	int count=1;
	for(i=0; i<probNn; i++){
		stack[i]=0;
		visited[i]=0;
	}
	//if(tid==0){
	//	for(i=0; i<probNl; i++){
	//		cuPrintf("%d",getBit(i,lbit));
	//	}
	//	cuPrintf("\n");
	//}
	stack[sp]=0;
	visited[0]=1;
	while(sp<count && count<probNn){
		for(i=0; i<probNn; i++){
			if(!visited[i]){
				eidx=edgeIndex(stack[sp],i);
				//if(tid==0)
				//	cuPrintf("[%d]=(%d,%d)",eidx,stack[sp],i);
				if(getBit(eidx,lbit)==1){
					stack[count++]=i;
					visited[i]=1;
					//if(tid==0)
					//	cuPrintf("%d",i);
				}
			}
		}
		sp++;
		//if(tid==0)
		//	cuPrintf("\nsp=%d\n",sp);
	}
	//if(tid==0)
	//	cuPrintf("\n");
	return count==probNn;
}

/*validno je ukoliko je povezano*/
__device__ int valid(unsigned char *lbit, int tid){
	return bfs(lbit,tid);
}


/*funckija za validnost koja dodatno ubacuje vestacki element
algoritamske slozenosti. Ubacuje se redudantan hod slozenosti O(n^P)
gde je P dato kao argument*/
__device__ int Pvalid(unsigned char *lbit, int P, int tid){
	if(P>0){
		int complexity=1;
		for(int p=0; p<P;p++)
			complexity*=probNn;
		int sum=0;
		for(int i=0; i<complexity; i++){
			if(k_rand(tid)%2==0)
				sum+=(k_rand(tid)%100);
			else
				sum-=(k_rand(tid)%100);
		}
	}
	return valid(lbit, tid);
}


/*poredi stringove istih dimenzija*/
__device__ int cuStrcmp(char *s1, char *s2){
	for(; *s1 == *s2; ++s1, ++s2)
		if(*s1 == 0)
			return 0;
	return *(unsigned char *)s1 < *(unsigned char *)s2 ? -1 : 1;
}

/***************************************************************\
selection()

Vrsi selekciju tako sto pronalazi najbolji medju NCOMPET slucajno
izabranih jedinki. Ideja je da se elitne jedinke ostavljaju. 
Nakon toga vrsi se odabir POPSIZE-NELITE jedinki turnirskom 
selekcijom. Na ovaj nacin, odredjene jedinke ce najv. imati
vise od 1 ponavljanja. 
\***************************************************************/
__device__ void selection(int tid, unsigned short int sfit[POPSIZE], unsigned char slbit[POPSIZE][EDGEMEMNUM])
{
	int intItem, intChosen, intTemp;
	int intChamp, intCompet;
	int indSelected=0;
	unsigned short int fitTemp;
	unsigned char lbitTemp[EDGEMEMNUM];

	intItem=tid;
	if(intItem>=NELITE && intItem<POPSIZE){
		intChamp= k_rand(intItem)%(POPSIZE-NELITE)+NELITE;
		for ( intChosen=1; intChosen<NCOMPET; intChosen++ ){
			intCompet= k_rand(intItem)%(POPSIZE-NELITE)+NELITE;
			if ((MINIMIZATION && sfit[intCompet]<sfit[intChamp])
				|| (!MINIMIZATION && sfit[intCompet]>sfit[intChamp]))
				intChamp= intCompet;
		}
		indSelected=intChamp;
		fitTemp=sfit[indSelected];
		for(intTemp=0; intTemp<probNl; intTemp++){
			setBit(intTemp,lbitTemp,getBit(intTemp,slbit[indSelected]));
		}
		__syncthreads();
		//sada kada su svi dobili lokacije, treba prekopirati te vrednosti
		sfit[intItem]=fitTemp;
		for(intTemp=0; intTemp<probNl; intTemp++){
			setBit(intTemp,slbit[intItem],getBit(intTemp,lbitTemp));
		}
	}
} 

__device__ void crossover(int tid, unsigned short int sfit[POPSIZE], unsigned char slbit[POPSIZE][EDGEMEMNUM])
{
	int i,j,itm1,itm2,site, i1, i2,temp;
	float crossProb;
	//pomocni niz za obelezavanje koji je hromozom vec
	//ucestvovao u rekombinaciji
	short int ind[POPSIZE];
	int popsize=POPSIZE;

	for(i=0; i<POPSIZE; i++)
		ind[i]=0;
	//za sve NEelitne jedinke vrsimo ukrstanje
	if(tid>=NELITE && tid%2==0){
		i=tid;
		//biramo slucajnu tacku ukrstanja
		//razmenjujemo kodove
		i1 = k_rand(i)%(popsize-NELITE);
		itm1 = NELITE;
		//biramo onu jedinku koja je prva najbliza
		//indeksu i1, a pritom nad njom nije vrsena
		//rekombinacija
		for(j=0; j<=i1;j++)
		{
			while (ind[itm1]) itm1++;
			itm1++;
		}
		itm1--;
		popsize--;
		ind[itm1] = 1;

		i2 = k_rand(i)%(popsize-NELITE);
		itm2 = NELITE;
		for(j=0; j<=i2;j++)
		{
			while (ind[itm2]) itm2++;
			itm2++;
		}
		itm2--;
		popsize--;
		ind[itm2] = 1;

		crossProb=k_rand(i)*1.0/RAND_MAX;
		if(crossProb<CROSSOVERPROB){
			//jednopoziciona rekombinacija
			site=k_rand(i)%probNl;
			for(j=site; j<probNl; j++){
				temp=getBit(j,slbit[itm1]);
				setBit(j,slbit[itm1],getBit(j,slbit[itm2]));
				setBit(j,slbit[itm2],temp);
			}
		}
	}
} 

/* thread bazirana varijanta mutacije*/
__device__ void mutation(int tid, unsigned char slbit[POPSIZE][EDGEMEMNUM],float cModifiedMutationProb) 
{
	int j;
	float random;
	int b;
	/* Ne koristimo frozen bitove, jer nisu podrzane atomicke operacije nad deljeom memorijom na ovoj kartici
	__shared__ int frozen[MAXEDGES];

	if(tid<POPSIZE){
	for(j=0; j<probNl; j++)
	frozen[j]=0;

	for(j=0; j<probNl; j++)
	frozen[j]+=getBit(j,slbit[tid]);
	}
	__syncthreads();
	*/
	if(tid>=NELITE && tid<POPSIZE){
		for(j=0; j<probNl; j++){
			random=k_rand(tid)*1.0/RAND_MAX;
			if(random<cModifiedMutationProb){
				b=getBit(j,slbit[tid]);
				//i sa verov. mutiranja je primenjujemo
				if (b){
					setBit(j,slbit[tid],0);
				}else{
					setBit(j,slbit[tid],1);
				}
			}
		}
	}
}



__host__ char * strrstr(const char *str, const char *strSearch){
	char *ptr, *last=NULL;
	ptr = (char*)str;
	while((ptr = strstr(ptr, strSearch))) last = ptr++;
	return last;
}

__device__ int k_rand(int i)
{
	step_next(i);
	return ((unsigned int) (next[i] / 0x10000) & 0x7FFF);
	return 1;
}

__global__ void k_srand(unsigned long int seeds[POPSIZE])
{
	int i;
	for(i=0; i<POPSIZE; i++){
		next[i]=seeds[i];
	}
}

/*ova implementacija radi samo ako je broj elemenata stepen broja 2
TODO: ovo se mora uraditi, jer se ne sortira dobro za pop od 384, koje koristimo
u eksperimentima*/
__device__  static void bitonicSort(int tid, unsigned short int sfit[POPSIZE], unsigned char slbit[POPSIZE][EDGEMEMNUM])
{
	unsigned short int shTemp;
	unsigned char charTemp;

	for (int k = 2; k <= POPSIZE; k *= 2)
	{
		for (int j = k / 2; j>0; j /= 2)
		{
			int ixj = tid ^ j;

			if (ixj > tid)
			{
				if ((tid & k) == 0)
				{
					if (sfit[tid] > sfit[ixj])
					{
						//cuPrintf("Swaping %d with %d\n",sfit[tid],sfit[ixj]);
						shTemp=sfit[tid];
						sfit[tid]=sfit[ixj];
						sfit[ixj]=shTemp;
						for(int i=0; i<EDGEMEMNUM; i++){
							charTemp=slbit[tid][i];
							slbit[tid][i]=slbit[ixj][i];
							slbit[ixj][i]=charTemp;
						}
					}
				}
				else
				{
					if (sfit[tid] < sfit[ixj])
					{
						shTemp=sfit[tid];
						sfit[tid]=sfit[ixj];
						sfit[ixj]=shTemp;
						for(int i=0; i<EDGEMEMNUM; i++){
							charTemp=slbit[tid][i];
							slbit[tid][i]=slbit[ixj][i];
							slbit[ixj][i]=charTemp;
						}
					}
				}
			}
			__syncthreads();
		}
	}
}


__device__ void replaceEqual(int tid, unsigned short int sfit[POPSIZE],unsigned char slbit[POPSIZE][EDGEMEMNUM]){

	if(tid<POPSIZE-1 && sfit[tid]==sfit[tid+1]){
			/*zamenjujemo drugi ako imaju isti fit. value.*/
			generateRandom(slbit[tid+1],tid+1);
			sfit[tid+1]=fitness(slbit[tid+1],tid);
	}
}


/*pretpostavlja se da je niz sa long int-ovima dovoljno velik da 
primi trazeni indeks
ako podesavamo 1 onda radimo disjunkciju sa nizom koji ima 
sve 0 i 1 na zeljenom mestu
inace radimo konjukciju sa komplamentom opisanog bita*/
__device__ void setBit(int idx,unsigned char *bitArr, int val){
	if(val==1){
		bitArr[idx / EDGEMEMSIZE]|=(1<<(idx%EDGEMEMSIZE));
	}
	else if(val==0){
		bitArr[idx / EDGEMEMSIZE]&=(~(1<<(idx%EDGEMEMSIZE)));
	}
}

__device__ int getBit(int idx,unsigned char *bitArr){
	return (bitArr[idx/EDGEMEMSIZE]&(1<<(idx%EDGEMEMSIZE)))>0;
}

__global__ void GAallInOneOperator(chrom *popcurrent, float *cModifiedMutationProb, unsigned short int *cBest){
	int i,j;
	int tid=threadIdx.x;
	/*kalkulacija za koriscenje deljene memorije:
	CUDA dozvoljava 16KB memorije po multi-procesoru, ovde imamo 1 multiprocesor, dakle treba se
	u tu kolicinu uklopiti. Ideja je da podelimo memoriju na sledece delove. Ako je npr. velicina
	populacije hromozoma 128, maksimalan broj grana 25*24/2=300 (300/8+2=40 bajtova za opis)
	i pod pretp. da je sizeof(int)=4, sizeof(short int)=2, sizeof(char)=1 imamo:
	int sfit[POPSIZE]					256*4=1024B
	char slbit[POPSIZE][(EDGEMEMNUM]	256*40=10240B
	*/
	/*deljena memorija za originale i kopije populacije*/
	__shared__ unsigned  short int sfit[POPSIZE];
	__shared__ unsigned char slbit[POPSIZE][EDGEMEMNUM];
	//pomocne promenljive
	int loc;
	float random;
	int shTemp;

	//cuPrintf("Na ulazu popcurrent[%d]=%d\n",tid,popcurrent[tid].fit);
	//prepisujemo vrednosti u deljenu memoriju
	if(tid<POPSIZE){
		sfit[tid]=popcurrent[tid].fit;
		for(i=0; i<EDGEMEMNUM; i++){
			slbit[tid][i]=popcurrent[tid].lbit[i];
		}
	}
	//cekamo da to urade sve niti
	__syncthreads();

	//da bi smanjili broj pristupa globalnoj memoriji u kernel smo ubacili petlju
	for(int p=0; p<KERNELLOOP; p++){
		selection(tid, sfit,slbit);
		__syncthreads();
		crossover(tid,sfit,slbit);
		__syncthreads();
		mutation(tid,slbit,*cModifiedMutationProb);
		__syncthreads();
		sfit[tid]=fitness(slbit[tid],tid);
		//if(tid==0)
			//cuPrintf("Dobijeni fitnes %d\n",sfit[tid]);
		//cekamo da svi hromozomi najdu f-je prilagodjenosti
		__syncthreads();
		//teko onda sortiranje
		bitonicSort(tid, sfit, slbit);
		__syncthreads();
		//if(tid==0)
		//	cuPrintf("Posle prvog sorta %d=>%d\n",tid,sfit[tid]);	
		//uklanjanje duplikata
		replaceEqual(tid,sfit,slbit);
		__syncthreads();
		//if(tid==0)
		//	cuPrintf("Posle replace %d\n",sfit[tid]);
		//ponovo soritramo 
		bitonicSort(tid, sfit, slbit);
		__syncthreads();
		//if(tid==0)
		//	cuPrintf("Posle drugog sorta %d=>%d\n",tid,sfit[tid]);	
	}
	//sfit[tid]=fitness(slbit[tid],tid);
	if(tid==0){
		//cuPrintf("Vrednost najbolje %d\n",sfit[tid]);
		//cuPrintf("Bitovi: ");
		//	for(i=0; i<probNl; i++){
		//		cuPrintf("%d",getBit(i,slbit[tid]));
		//	}
		//	cuPrintf("\n");
	}
	//cuPrintf("%d=>%d\n",tid,sfit[tid]);
	//vracamo nazad u globalnu memoriju
	popcurrent[tid].fit=sfit[tid];
	for(j=0; j<probNl; j++){
		setBit(j,popcurrent[tid].lbit,getBit(j,slbit[tid]));
	}
	if(tid==0)
		*cBest=sfit[0];
	__syncthreads();
}

/************************************************************************/
/* Init CUDA                                                            */
/************************************************************************/
#if __DEVICE_EMULATION__

bool InitCUDA(void){return true;}

#else
bool InitCUDA(void)
{
	int count = 0;
	int i = 0;

	cudaGetDeviceCount(&count);
	if(count == 0) {
		fprintf(stderr, "There is no device.\n");
		return false;
	}

	for(i = 0; i < count; i++) {
		cudaDeviceProp prop;
		if(cudaGetDeviceProperties(&prop, i) == cudaSuccess) {
			if(prop.major >= 1) {
				break;
			}
		}
	}
	if(i == count) {
		fprintf(stderr, "There is no device supporting CUDA.\n");
		return false;
	}
	cudaSetDevice(i);

	printf("CUDA initialized.\n");
	return true;
}

#endif


void cudaDevicePropSummary(){
	cudaDeviceProp prop;
	int count;
	HANDLE_ERROR( cudaGetDeviceCount( &count ) );
	for (int i=0; i< count; i++) {
		HANDLE_ERROR( cudaGetDeviceProperties( &prop, i ) );
		printf( " --- General Information for device %d ---\n", i );
		printf( "Name: %s\n", prop.name );
		printf( "Compute capability: %d.%d\n", prop.major, prop.minor );
		printf( "Clock rate: %d\n", prop.clockRate );
		printf( "Device copy overlap: " );
		if (prop.deviceOverlap)
			printf( "Enabled\n" );
		else
			printf( "Disabled\n" );
		printf( "Kernel execition timeout : " );
		if (prop.kernelExecTimeoutEnabled)
			printf( "Enabled\n" );
		else
			printf( "Disabled\n" );
		printf( " --- Memory Information for device %d ---\n", i );
		printf( "Total global mem: %ld\n", prop.totalGlobalMem );
		printf( "Total constant Mem: %ld\n", prop.totalConstMem );
		printf( "Max mem pitch: %ld\n", prop.memPitch );
		printf( "Texture Alignment: %ld\n", prop.textureAlignment );
		printf( " --- MP Information for device %d ---\n", i );
		printf( "Multiprocessor count: %d\n",
			prop.multiProcessorCount );
		printf( "Shared mem per mp: %ld\n", prop.sharedMemPerBlock );
		printf( "Registers per mp: %d\n", prop.regsPerBlock );
		printf( "Threads in warp: %d\n", prop.warpSize );
		printf( "Max threads per block: %d\n",
			prop.maxThreadsPerBlock );
		printf( "Max thread dimensions: (%d, %d, %d)\n",
			prop.maxThreadsDim[0], prop.maxThreadsDim[1],
			prop.maxThreadsDim[2] );
		printf( "Max grid dimensions: (%d, %d, %d)\n",
			prop.maxGridSize[0], prop.maxGridSize[1],
			prop.maxGridSize[2] );
		printf( "\n" );
	}
}


int main(int argc, char* argv[])
{
	unsigned int randSeed;
	int numOfGen;
	int objhardness;
	float grphdens;
	cudaDevicePropSummary();

#ifndef CUDA_NO_SM_11_ATOMIC_INTRINSICS
	printf("WARNING! Not using atomics!\n");
#endif


	if(argc>5){
		randSeed=atoi(argv[2]);
		numOfGen=atoi(argv[3]);
		objhardness=atoi(argv[4]);
		grphdens=atof(argv[5]);
		ga(argv[1],randSeed,numOfGen,objhardness,grphdens);
	}else{
		fprintf(stderr, "Upotreba: <prog.exe> <ulFajl> <randSeed> <brGeneracija> <tezinaFjeCilja [0-N]> <gustinaGrafa [0,1]>");
		exit(0);
	}

	return 0;
}
